#include "enumeration.h"
#include <algorithm>

bool same(double x, double y) {
	if (fabs(x - y) < SMALL)
		return true;
	return false;
}



void normalize(double centerx, double centery, double ptx, double pty, double* nrmptx, double* nrmpty){
    *nrmptx = ptx - centerx;
    *nrmpty = pty - centery;
}

int quarter_of_point(double x, double y){
    //Clockwise, 12 to 3 is Q1, 3 to 6 is Q2, 6 to 9, Q3 and 9 to 12 is Q4
    if(y >= SMALL && x > -SMALL)
        return 1;
    if(y < SMALL && x >= SMALL)
        return 2;
    if(y <= -SMALL && x < SMALL)
        return 3;
    if(y > -SMALL && x <= -SMALL)
        return 4;
	printf("Should not find myself here. Abort.\n");
	std::abort();
	return -1;//To avoid compiler warning
    
}

void Enumerate::clockwise_points(double x, double y, int two_t){
    //x and y are center. Given two_t points around (x,y)
    int noq1pts = 0;//Includes +y axis, excludes +x axis
    int noq2pts = 0;//Includes +x axis, excludes -y axis
    int noq3pts = 0;//Includes -y axis, excludes -x axis
    int noq4pts = 0;//Includes -x axis, excludes +y axis
    double norm_x, norm_y;
    for(int ptindx = 0; ptindx < two_t; ptindx++){
        normalize(x, y, TNbrIPs[ptindx].first, TNbrIPs[ptindx].second, &norm_x, &norm_y);
        int qrtr = quarter_of_point(norm_x, norm_y);
        if(qrtr == 1){
            if(same(0, norm_x))
                Qslopes[ptindx] = LARGED;
            else
				Qslopes[ptindx] = norm_y / norm_x;
            TNbrIPsIndexClkWiseQ1[noq1pts] = ptindx;
            noq1pts++;
        }
        else{
            if(qrtr == 2){
                if(same(0, norm_y))
					Qslopes[ptindx] = LARGED;
                else
					Qslopes[ptindx] = norm_x / -norm_y;
                TNbrIPsIndexClkWiseQ2[noq2pts] = ptindx;
                noq2pts++;
            }
            else{
                if(qrtr == 3){
                    if(same(0, norm_x))
						Qslopes[ptindx] = LARGED;
                    else
						Qslopes[ptindx] = norm_y / norm_x; 
                    TNbrIPsIndexClkWiseQ3[noq3pts] = ptindx;
                    noq3pts++;
                }
                else{
                    if(same(0, norm_y))
						Qslopes[ptindx] = LARGED;
                    else
						Qslopes[ptindx] = -norm_x / norm_y;
                    TNbrIPsIndexClkWiseQ4[noq4pts] = ptindx;
                    noq4pts++;
                }
            }
        }
    }
    
    std::sort(TNbrIPsIndexClkWiseQ1.begin(), TNbrIPsIndexClkWiseQ1.begin() + noq1pts, [&](int index_left, int index_right) { return Qslopes[index_left] > Qslopes[index_right]; });//Descending order
    std::sort(TNbrIPsIndexClkWiseQ2.begin(), TNbrIPsIndexClkWiseQ2.begin() + noq2pts, [&](int index_left, int index_right) { return Qslopes[index_left] > Qslopes[index_right]; });//Descending order
    std::sort(TNbrIPsIndexClkWiseQ3.begin(), TNbrIPsIndexClkWiseQ3.begin() + noq3pts, [&](int index_left, int index_right) { return Qslopes[index_left] > Qslopes[index_right]; });//Descending order
	std::sort(TNbrIPsIndexClkWiseQ4.begin(), TNbrIPsIndexClkWiseQ4.begin() + noq4pts, [&](int index_left, int index_right) { return Qslopes[index_left] > Qslopes[index_right]; });//Descending order
    
    int runningindex = 0;
	printf("Points in quarter 1: ");
    for(int i = 0; i < noq1pts; i++){
        TNbrIPsIndexClkWise[runningindex] = TNbrIPsIndexClkWiseQ1[i];
		printf("(%f, %f) ", TNbrIPs[TNbrIPsIndexClkWise[runningindex]].first, TNbrIPs[TNbrIPsIndexClkWise[runningindex]].second);
        runningindex++;
    }
	printf("\n");
	printf("Points in quarter 2: ");
    for(int i = 0; i < noq2pts; i++){
        TNbrIPsIndexClkWise[runningindex] = TNbrIPsIndexClkWiseQ2[i];
		printf("(%f, %f) ", TNbrIPs[TNbrIPsIndexClkWise[runningindex]].first, TNbrIPs[TNbrIPsIndexClkWise[runningindex]].second);
        runningindex++;
    }
	printf("\n");
	printf("Points in quarter 3: ");
    for(int i = 0; i < noq3pts; i++){
        TNbrIPsIndexClkWise[runningindex] = TNbrIPsIndexClkWiseQ3[i];
		printf("(%f, %f) ", TNbrIPs[TNbrIPsIndexClkWise[runningindex]].first, TNbrIPs[TNbrIPsIndexClkWise[runningindex]].second);
        runningindex++;
    }
	printf("\n");
	printf("Points in quarter 4: ");
    for(int i = 0; i < noq4pts; i++){
        TNbrIPsIndexClkWise[runningindex] = TNbrIPsIndexClkWiseQ4[i];
		printf("(%f, %f) ", TNbrIPs[TNbrIPsIndexClkWise[runningindex]].first, TNbrIPs[TNbrIPsIndexClkWise[runningindex]].second);
        runningindex++;
    }
	printf("\n");
}

double euclidean_distance(double x1, double y1, double x2, double y2) {
	return sqrt ((   (x1 - x2)*(x1 - x2) )        +		((y1 - y2) * (y1 - y2))				);
}


void Enumerate::findip(struct circle_s* c1, struct circle_s* c2) {

	double k = pow(c2->radius, 2.) - pow(c1->radius, 2.) + pow(c1->x, 2.) - pow(c2->x, 2.);
	k /= 2. * (c1->x - c2->x);
	//Below are coefficients of equation quadratic in y, Ay2 + By + C = 0
	//where it is known that the circles intersect and there are two intersection points and the y coordinates of the centers' are the same.
	//More general cases need to be considered separately.
	double A = 1.;
	double B = -2. * c1->y;
	double C = k * k - 2. * k * c1->x - c1->radius * c1->radius + c1->x * c1->x + c1->y * c1->y;
	double determinant = pow(B, 2.) - 4. * A * C;
	if (determinant <= -SMALL || determinant < SMALL) {
		printf("Either intersection is at one singular point, or there is no intersection at all. Those cases need to be considered separately.\n");
		getchar();
		std::abort();
	}
	double ystar = (-B + sqrt(determinant)) / 2.0 / A;
	double xstar = k;
	//Store the intersection point in ips[]
	ips.emplace_back();
	ips.back().i = c1->customer;
	ips.back().j = c2->customer;
	ips.back().x = xstar;
	ips.back().y = ystar;



	ystar = (-B - sqrt(determinant)) / 2.0 / A;
	ips.emplace_back();
	ips.back().i = c1->customer;
	ips.back().j = c2->customer;
	ips.back().x = xstar;
	ips.back().y = ystar;

}

double min(double x, double y) {
	if (x < y)
		return x;
	return y;
}


void Enumerate::find_circles_intersection(int cust1, double x2, double y2, double r2, int tindex) {
	//For circle corresponding to 'tindex' in std::vector T, cust1, given neighbourhood with radius r2 centered at (x2, y2)
	//you know for sure that there will be two intersection points of this neighbourhood with
	//circle corresponding to tindex. Find those points. In Figure (3) of the paper, for 'C1' and 'Nbr', this is points 8 and 4.
	double x1 = circles[cust1].x;
	double y1 = circles[cust1].y;
	double r1 = circles[cust1].radius;
	if (fabs(x2 - x1) >= SMALL) {//The points' X coordinates are different
		double kden = 2. * (x1 - x2);
		double kcnst = (pow(x1, 2) - pow(x2, 2) + pow(y1, 2) - pow(y2, 2)) / kden;
		double k = kcnst + (pow(r2, 2.) - pow(r1, 2.)) / kden;
		double dbac = (y2 - y1) / (x1 - x2);
		double Acnst = 1. + pow(dbac, 2.);
		double Bcnst = -2. * dbac * x1 - 2. * y1;
		double Ccnst = pow(x1, 2.) + pow(y1, 2.);
		double A = Acnst;
		double B = Bcnst + 2. * k * dbac;
		double C = Ccnst + k * k - 2. * k * x1 - r1 * r1;
		double determinant = pow(B, 2.) - 4. * A * C;
		if (determinant <= -SMALL || determinant < SMALL) {
			//For sure, given a neighbourhood, it will intersect the circle at two distinct points.
			//So, this branching should not be entered into
			printf("Should not be here.");
			getchar();
			std::abort();
		}
		double ystar = (-B + sqrt(determinant)) / 2.0 / A;
		double xstar = k + ystar * dbac;

		TNbrIPs[2 * tindex].first = xstar;
		TNbrIPs[2 * tindex].second = ystar;

		ystar = (-B - sqrt(determinant)) / 2.0 / A;
		xstar = k + ystar * dbac;

		TNbrIPs[2 * tindex + 1].first = xstar;
		TNbrIPs[2 * tindex + 1].second = ystar;


	}
	else {
		//Points have same x coordinates.
		double kden = 2. * (y1 - y2);
		double kcnst = (pow(y1, 2) - pow(y2, 2)) / kden;
		double Acnst = 1.;
		double Bcnst = -2. * x1;
		double Ccnst = pow(x1, 2.) + pow(y1, 2.);
		double k = kcnst + (pow(r2, 2.) - pow(r1, 2.)) / kden;
		double A = 1.;
		double B = Bcnst;
		double C = Ccnst + k * k - 2. * k * y1 - r1 * r1;
		double determinant = pow(B, 2.) - 4. * A * C;
		if (determinant <= -SMALL || determinant < SMALL) {
			//For sure, given a neighbourhood, it will intersect the circle at two distinct points.	
			//So, this branching should not be entered into
			printf("Should not be here.");
			getchar();
			std::abort();
		}
		double xstar = (-B + sqrt(determinant)) / 2.0 / A;
		double ystar = k;

		TNbrIPs[2 * tindex].first = xstar;
		TNbrIPs[2 * tindex].second = ystar;

		xstar = (-B - sqrt(determinant)) / 2.0 / A;
		TNbrIPs[2 * tindex + 1].first = xstar;
		TNbrIPs[2 * tindex + 1].second = ystar;
	}
}

bool does_circle_contain_point(double ctr_x, double ctr_y, double rad, double ptx, double pty) {
	double dist = euclidean_distance(ctr_x, ctr_y, ptx, pty);
	if (dist < rad + SMALL)
		return true;
	return false;
}

void Enumerate::enumerate_regions() {

	//Create space for TNbrIPs datastructure that will store the intersection points of the sufficiently small neighborhood 
	//with each of the circles that intersect at the center of the neighborhood. Need space for 2N points
	for (int i = 0; i < 2 * N; i++)
		TNbrIPs.push_back(std::make_pair<double, double>(0, 0));
	//Now create space for other datastructures that are going to be used to traverse these points in clockwise order.
	for (int i = 0; i < N; i++) {
		TNbrIPsIndexClkWise.push_back(0);
		TNbrIPsIndexClkWise.push_back(0);
		TNbrIPsIndexClkWiseQ1.push_back(0);
		TNbrIPsIndexClkWiseQ1.push_back(0);
		TNbrIPsIndexClkWiseQ2.push_back(0);
		TNbrIPsIndexClkWiseQ2.push_back(0);
		TNbrIPsIndexClkWiseQ3.push_back(0);
		TNbrIPsIndexClkWiseQ3.push_back(0);
		TNbrIPsIndexClkWiseQ4.push_back(0);
		TNbrIPsIndexClkWiseQ4.push_back(0);
		Qslopes.push_back(0);
		Qslopes.push_back(0);
	}

	struct circle_s* circle1;
	struct circle_s* circle2;
	for (int c1 = 0; c1 < N; c1++) {
		circle1 = &circles[c1];
		for (int c2 = c1 + 1; c2 < N; c2++) {
			circle2 = &circles[c2];
			findip(circle1, circle2);
		}
	}
	for (int ipindex = 0, sz = static_cast<int>(ips.size()); ipindex < sz; ipindex++) {
		double x = ips[ipindex].x;
		double y = ips[ipindex].y;
		std::bitset<N> btst_S;//Indicator vector of set of circles that contain (x,y) in their interior
		std::bitset<N> btst_T;//Indicator vector of set of circles that contain (x,y) on their boundary
		for (int cust = 0; cust < N; cust++) {
			if (cust == ips[ipindex].i || cust == ips[ipindex].j) {//If cust contributed to this intersection point, load it into T automatically
				btst_T.set(cust);//T will contain the original contribution pair of customers.
				continue;
			}

			double dist = euclidean_distance(x, y, circles[cust].x, circles[cust].y);//Distance between (x,y) and customer cust
			if (circles[cust].radius >= dist + SMALL) {
				//Circle around cust exceeds distance between IP and cust. Hence, the circle encompasses the IP
				btst_S.set(cust);
			}
			else {
				//Circle does NOT encompass the IP.
				if (circles[cust].radius > dist - SMALL) {
					btst_T.set(cust);
				}
			}
		}
		//Load set S's into a std::vector
		std::vector<int> S;
		S.reserve(N);
		for (int i = 0; i < N; i++) {
			if (btst_S.test(i) == true)
				S.emplace_back(i);
		}
		//Load set T's into a std::vector
		std::vector<int> T;
		T.reserve(N);
		for (int i = 0; i < N; i++) {
			if (btst_T.test(i) == true)
				T.emplace_back(i);
		}
		int s = static_cast<int>(S.size());
		int t = static_cast<int>(T.size());
		//At this stage, if t = 1, T = {C1}, say
		//regions to enumerate are C1 and emptyset.
		//That case needs to be considered separately

		
		double nbr_rad = LARGED;//Need to find a neighbourhood sufficiently small



		for (int sindex = 0; sindex < s; sindex++) {
			int scust = S[sindex];
			double temprad = circles[scust].radius;
			temprad -= euclidean_distance(x, y, circles[scust].x, circles[scust].y);
			nbr_rad = min(nbr_rad, temprad);
		}

		for (int tindex1 = 0; tindex1 < t; tindex1++) {
			int tcust1 = T[tindex1];
			nbr_rad = min(nbr_rad, 2. * circles[tcust1].radius);
			for (int tindex2 = tindex1 + 1; tindex2 < t; tindex2++) {
				int tcust2 = T[tindex2];
				//Below is specialized for case where y coordinates of circle centers are the same. 
				//The general goal is that (reflectedx, reflectedy) is the reflection about the line joining the centers of circles tcust1 and tcust2
				//of the other intersection point, (x,y)
				double a = 1.;
				double b = 0.;
				double c = -circles[tcust1].y;
				double denominator = a * a + b * b;
				double a2mb2 = a * a - b * b;
				double reflectedx = (x * a2mb2 - 2. * b * (a * y + c)) / denominator;
				double reflectedy = (y * (-a2mb2) - 2. * a * (b * x + c)) / denominator;
				double temprad = euclidean_distance(x, y, reflectedx, reflectedy);
				if (temprad >= SMALL) {//There is a second intersection point
					nbr_rad = min(nbr_rad, temprad);
				}
			}
		}
		nbr_rad /= 2.;
		if (nbr_rad < EXHENUMERATIONTHRESHOLD + SMALL) {
			//In some cases, nbr_rad can be exceedingly small, yet positive, due to numerical issues.
			//In this case, there is no choice but to exhaustively enumerate 2^|T| regions.
			//Many of these regions can subsequently be processed and be seen to be infeasible.
			printf("Exhaustive Enumeration at %f\n", nbr_rad);
			int n = static_cast<int>(T.size());
			int count = static_cast<int>(pow(2, n));
			for (int i = 0; i < count; i++) {
				std::bitset<N> start;//ALL FALSE
				//Loop below creates the binary representation of integer i
				for (int j = 0; j < n; j++)
				{
					if ((i & (1 << j)) > 0)
						start.set(T[j]);
				}
				TClusters_btst.insert(start);
			}
		}
		else {
			for (int i = 0; i < t; i++)
				find_circles_intersection(T[i], x, y, nbr_rad, i);
			clockwise_points(x, y, 2 * (t));
			//Number of intersection points will be 2t 
			std::bitset<N> prev;
			std::bitset<N> current;
			for (int i = 0; i < 2 * t; i++) {
				int indexinto = TNbrIPsIndexClkWise[i];
				double ipx = TNbrIPs[indexinto].first;
				double ipy = TNbrIPs[indexinto].second;
				if (i > 0) {
					prev = current;
				}
				current.reset();
				current[ips[ipindex].i] = does_circle_contain_point(circles[ips[ipindex].i].x, circles[ips[ipindex].i].y, circles[ips[ipindex].i].radius, ipx, ipy);
				current[ips[ipindex].j] = does_circle_contain_point(circles[ips[ipindex].j].x, circles[ips[ipindex].j].y, circles[ips[ipindex].j].radius, ipx, ipy);
				for (int crcindx = 0; crcindx < t; crcindx++) {
					current[T[crcindx]] = does_circle_contain_point(circles[T[crcindx]].x, circles[T[crcindx]].y, circles[T[crcindx]].radius, ipx, ipy);
				}
				TClusters_btst.insert(current);
				if (i > 0) {
					//Now take intersection between current and prev.
					for (int bt = 0; bt < N; bt++) {
						if (prev[bt] == true && current[bt] == true)
							current[bt] = true;
						else
							current[bt] = false;
					}
					TClusters_btst.insert(current);
				}
			}
		}

		//Now, print out the enumerated subsets of T
		for (std::unordered_set<std::bitset<N>>::iterator it = TClusters_btst.begin(); it != TClusters_btst.end(); it++) {
			std::bitset<N> btst = btst_S;
			for (int i = 0; i < N; i++) {
				if ((*it)[i] == true) {
					btst.set(i);
				}
			}
			if (btst.any()) {
				//should be nonempty
				printf("For intersection point %d (%f, %f) between circles %d and %d\n", ipindex, ips[ipindex].x, ips[ipindex].y, ips[ipindex].i, ips[ipindex].j);
				for (int i = 0; i < N; i++) {
					if (btst[i])
						printf("%d ", i);
				}
				printf("\n");
			}
		}
	}
}
